module RuneRb::System::Types
  # A Stack object organizes work via RuneRb::System::Types::Task objects and ensures the sequential execution of Tasks based on their Priority.
  class SortedStack

    # Constructs a new Stack object
    # @return [Stack]
    def initialize
      @registry = []
    end

    # Pushes a new task to the Stack. Options are passed directly to the Task constructor.
    #
    # The <@registry> object is a SortedSet which automatically organizes each task on submission and links the task's process to its predecessor's (via Task#target_to).
    # @param task [RuneRb::System::Types::Task] the task to post.
    def push(task)
      @registry << task
    end

    alias << push

    # Returns whether the <@registry> is empty or not via boolean values (true, false)
    # @return [Boolean] is the <@registry> empty?
    def empty?
      @registry.empty?
    end

    # Returns the first task in the <@registry>
    # @return [RuneRb::System::Types::Task, NilClass] the first task in the registry.
    def first
      @registry.first
    end

    # Returns the last task in the <@registry>
    # @return [RuneRb::System::Types::Task, NilClass] the last task in the registry.
    def last
      @registry.last
    end

    # Returns the length of the registry
    # @return [Integer]
    def length
      @registry.length
    end

    # Sorts the registry and links it's children via Task#target_to
    def sort!
      @registry.sort!
      @registry.each do |task|
        next if task == last

        task.target_to(@registry[@registry.index(task) + 1].worker)
      end
    end

    # Removes all Tasks from the <@registry>
    def clear
      @registry.clear
    end
  end
end

# Copyright (c) 2021, Patrick W.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.