module RuneRb::System::Types

  # A Task object encapsulates a Fiber that will execute code, then transfer execution context to a target Fiber.
  class Task
    include RuneRb::System::Utils::Logging

    # @!attribute [r] id
    # @return [Integer, Symbol] the Task ID.
    attr :id

    # @!attribute [r] priority
    # @return [Symbol] the Task priority.
    attr :priority

    # @!attribute [r] worker
    # @return [Fiber] the Fiber that will execute the Task.
    attr :worker

    # @!attribute [r] target
    # @return [Fiber] the target Fiber.
    attr :target

    # Constructs a new Task object.
    # @param params [Hash] Initial parameters for the Task.
    # @param _ [Proc] parameters passed to the block operations
    def initialize(params = {}, &_)
      @id = params[:id] || Druuid.gen
      @assets = params[:assets] || []
      @priority = params[:priority] || :LOW
      @worker = Fiber.new do |assets|
        loop do
          result = yield(assets)
        rescue StandardError => e
          err 'An error occurred while processing task!', e, e.message, e.backtrace&.join("\n")
          break
        ensure
          break if @completed

          @target ? @target.transfer(result) : Fiber.yield(result)
        end
      end
      start if params[:autorun]
    end

    # Calls <@worker#resume> causing it to process its work.
    def start
      @worker.resume(@assets)
    end

    # Sets the value of <@completed> to true, preventing further execution of the <@worker>.
    def stop
      @completed = true
      # We do this to ensure the loop completes
      start
    end

    # Update the <@target> Fiber to the passed object.
    # @param fiber [Fiber] the new target Fiber.
    # @todo create a InvalidTarget error to properly handle scenarios where a Fiber isn't supplied to this function.
    def target_to(fiber)
      raise "Invalid target; must be a Fiber! Received: #{fiber.class}" unless fiber.is_a?(Fiber)

      @target = fiber
    end

    def inspect
      "[id]: #{@id}\t||\t[Priority]: #{@priority}"
    end

    # Mutual comparison operator. Used to sort the Task by it's priority.
    # @param other [Task] the compared Task
    def <=>(other)
      RuneRb::System::TASK_PRIORITIES[@priority] <=> RuneRb::System::TASK_PRIORITIES[other.priority]
    end
  end
end

# Copyright (c) 2021, Patrick W.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.